REBOL [
	Title: "REBOL Mezzanine Functions: Control"
	Version: 2.7.8
	Rights: "Copyright REBOL Technologies 2009. All rights reserved."
	Home: http://www.rebol.com
	Date: 30-Jan-2010

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.

	; Portions of this code are licensed from the R2/Forward project.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

; Loop functions

forall: func [
	"Evaluates a block for every value in a series."
	[catch throw]
	'word [word!] {Word set to each position in series and changed as a result}
	body [block!] "Block to evaluate each time"
][
	throw-on-error [forskip :word 1 body]
]

forskip: func [
	"Evaluates a block for periodic values in a series."
	[throw catch]
	'word [word!] {Word set to each position in series and changed as a result}
	skip-num [integer!] "Number of values to skip each time"
	body [block!] "Block to evaluate each time"
	/local orig result
][
	if not positive? skip-num [throw make error! join [script invalid-arg] skip-num]
	if not any [
		series? get word
		port? get word
	] [throw make error! "forskip/forall expected word argument to refer to a series or port!"]
	orig: get word
	while [any [not tail? get word  (set word orig  false)]] [
		set/any 'result do body
		set word skip get word skip-num
		get/any 'result
	]
]

for: func [
	{Repeats a block over a range of values.}
	[catch throw]
	'word [word!] "Variable to hold current value"
	start [number! series! money! time! date! char!] "Starting value"
	end   [number! series! money! time! date! char!] "Ending value"
	bump  [number! money! time! char!] "Amount to skip each time"
	body  [block!] "Block to evaluate"
	/local result do-body op
][
	if (type? start) <> (type? end) [
		throw make error! reduce ['script 'expect-arg 'for 'end type? start]
	]
	do-body: func reduce [[ throw ] word] body
	op: :greater-or-equal?
	either series? start [
		if not same? head start head end [
			throw make error! reduce ['script 'invalid-arg end]
		]
		if (negative? bump) [op: :lesser?]
		while [op index? end index? start] [
			set/any 'result do-body start
			start: skip start bump  ; won't go back past head
		]
		if (negative? bump) [set/any 'result do-body start] ; need one more
	][
		if (negative? bump) [op: :lesser-or-equal?]
		while [op end start] [
			set/any 'result do-body start
			start: start + bump
		]
	]
	get/any 'result
]

forever: func [
	{Evaluates a block endlessly.}
	[throw]
	body [block!] {Block to evaluate each time}
][
	while [on] body
]


; Control functions

!: :not

++: func [
	"Increment an integer or series index. Return its prior value."
	[catch]
	'word [word! paren!] "Integer or series variable."
][
	if all [paren? word not word? set/any 'word do word] [
		throw-error 'script 'expect-arg reduce ['++ 'word type? get/any 'word]
	] ; Workaround for R3 change in lit-word! parameters with paren! arguments
	case [
		number? get/any word [also get word set word add get word 1]
		series? get/any word [also get word set word next get word]
		'else [throw-error 'script 'expect-arg reduce ['++ 'word type? get/any 'word]]
	]
]
; Note: Native in R3.

--: func [
	"Decrement an integer or series index. Return its prior value."
	[catch]
	'word [word! paren!] "Integer or series variable."
][
	if all [paren? word not word? set/any 'word do word] [
		throw-error 'script 'expect-arg reduce ['-- 'word type? get/any 'word]
	] ; Workaround for R3 change in lit-word! parameters with paren! arguments
	case [
		number? get/any word [also get word set word subtract get word 1]
		series? get/any word [also get word set word back get word]
		'else [throw-error 'script 'expect-arg reduce ['-- 'word type? get/any 'word]]
	]
]
; Note: Native in R3.

also: func [
	{Returns the first value, but also evaluates the second.}
	value1 [any-type!]
	value2 [any-type!]
][
	return get/any 'value1
]

quote: func [
	"Returns the value passed to it without evaluation."
	:value [any-type!]
][ ; Broken for word! arguments - use lit-word! instead.
	return get/any 'value
]
; Based on a proposal by Peta.

true?: func [
	"Returns true if an expression can be used as true."
	val
][not not :val]
; Code written by Carl, spec by BrianH.

default: func [
	"Set a word to a default value if it hasn't been set yet."
	'word [word! set-word! lit-word!] "The word (use :var for word! values)"
	value "The value" ; unset! not allowed on purpose
][
	unless all [value? word not none? get word] [set word :value] :value
]
; Suggested by a discussion in R3 chat in Jan-2009.

attempt: func [
	"Tries to evaluate and returns result or NONE on error."
	[throw]
	value ;[block!]  ; Unnecessary overhead - TRY tests for this type
][
	unless error? set/any 'value try :value [get/any 'value]
]

resolve: func [
	"Copy context by setting values in the target from those in the source."
	[catch]
	target [object! port!]
	source [object! port!]
	/only from [block! integer!] "Only specific words (exports) or new words in target (index to tail)"
	/all "Set all words, even those in the target that already have a value"
][
	either only [
		from: either integer? from [
			; Only set words in the target positioned at the number from or later
			unless positive? from [throw-error 'script 'out-of-range from]
			intersect words-of source at words-of target from
		] [
			; Only set the words in the target that are also in the from block
			intersect words-of source intersect words-of target from
		]
		foreach word from pick [
			[unless value? in target word [error? set/any in target word get/any word]]
			[error? set/any in target word get/any word]
		] not all ; See below for what this means
	] [
		either all [ ; Override all target words even if they have values
			error? set/any bind words-of source target get/any source
		] [ ; Only set target words if they aren't yet set
			foreach word intersect words-of source words-of target [
				unless value? in target word [error? set/any in target word get/any word]
			]
		]
	]
	also target set [source target from] none
]
; Note: This is native in R3 and supports module! too.
; Implementation note: INTERSECT returns the values from its first argument,
; and WORDS-OF returns a block of words that are bound to its argument, so
;     intersect words-of source words-of target
; returns a block of words bound to source.

; new switch from RAMBO #4143, with added /all
;switch: func [
;	"Selects a choice and evaluates the block that follows it."
;	[ throw ]
;	value "Value to search for."
;	cases [block!] "Block of cases to search."
;	/default case "Default case if no others are found."
;	/all "Evaluate all matches (not just first one)."
;
;	/local code found?
;][
;	code: clear [ ]
;	while [cases: find cases value] [
;		either cases: find next cases block! [
;			found?: yes
;			append code first cases
;			cases: next cases
;			unless all [break]
;		] [break]
;	]
;	do either found? [code] [case]
;]

dispatch: func [
	"Wait for a block of ports. As events happen, dispatch port handler blocks."
	port-block [block!] "Block of port handler pairs (port can be timeout too)."
	/local ports awake timeblk result
][
	ports: copy [] ;list of ports to wait
	foreach [port job] port-block: reduce port-block [
		if any [number? port time? port][if none? timeblk [timeblk: :job]]
		append ports port ; for a timeout value too
	]
	forever [
		either awake: wait/all ports [
			if foreach item awake [
				set/any 'result do select port-block item item ; arg is optional
				if all [value? 'result 'break = :result] [break/return true]
			][break]
		][do :timeblk]
	]
]
