REBOL [
	Title: "REBOL Mezzanine Functions: String"
	Version: 2.7.8
	Rights: "Copyright REBOL Technologies 2009. All rights reserved."
	Home: http://www.rebol.com
	Date: 30-Jan-2010

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.

	; Portions of this code are licensed from the R2/Forward project.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

; Character/string encoding functions

ascii?: funct [
	"Returns TRUE if value or string is in ASCII character range (below 128)."
	value [string! file! email! url! tag! issue! char! integer!]
] compose [
	ascii: (charset [#"^(00)" - #"^(7F)"])
	either any-string? value [parse/all/case value [any ascii]] [value < 128]
]
; Note: Native in R3.

latin1?: func [
	"Returns TRUE if value or string is in Latin-1 character range (below 256)."
	value [string! file! email! url! tag! issue! char! integer!] ; Not binary!
][ ; R2 has Latin-1 chars and strings
	either integer? value [value < 256] [true]
]
; Note: Native (and more meaningful) in R3. For forwards compatibility.

utf?: func [
	"Returns the UTF encoding from the BOM (byte order marker): + for BE; - for LE."
	data [binary!]
][
	parse/all/case data [
		#{EF BB BF} (return 8) |
		#{00 00 FE FF} (return 32) |
		#{FF FE 00 00} (return -32) |
		#{FE FF} (return 16) |
		#{FF FE} (return -16) |
		(return 0) ; No BOM or an unsupported BOM
	] ; Note: R3 only supports these BOMs.
]
; Note: Native in R3.

invalid-utf?: funct [
	"Checks for proper UTF encoding and returns NONE if correct or position where the error occurred."
	[catch]
	data [binary!]
	/utf "Check encodings other than UTF-8"
	num [integer!] "Bit size - positive for BE negative for LE"
] compose [
	ascii: (charset [#"^(00)" - #"^(7F)"])
	utf8+1: (charset [#"^(C2)" - #"^(DF)"])
	utf8+2: (charset [#"^(E0)" - #"^(EF)"])
	utf8+3: (charset [#"^(F0)" - #"^(F4)"])
	utf8rest: (charset [#"^(80)" - #"^(BF)"])
	switch/default any [num 8] [
		8 [ ; UTF-8
			unless parse/all/case data [(pos: none) any [
				pos: ascii | utf8+1 utf8rest |
				utf8+2 2 utf8rest | utf8+3 3 utf8rest
			]] [as-binary pos]
		]
		16 [ ; UTF-16BE
			pos: data
			while [not tail? pos] [
				hi: first pos
				case [
					none? lo: pick pos 2 [break/return pos]
					55296 > w: hi * 256 + lo [pos: skip pos 2]  ; #{D800}
					57343 < w [pos: skip pos 2]  ; #{DFFF}
					56319 < w [break/return pos]  ; #{DBFF}
					none? hi: pick pos 3 [break/return pos]
					none? lo: pick pos 4 [break/return pos]
					56320 > w: hi * 256 + lo [break/return pos]  ; #{DC00}
					57343 >= w [pos: skip pos 4]  ; #{DFFF}
				]
				none
			] ; none = valid, break/return pos = invalid
		]
		-16 [ ; UTF-16LE
			pos: data
			while [not tail? pos] [
				lo: first pos
				case [
					none? hi: pick pos 2 [break/return pos]
					55296 > w: hi * 256 + lo [pos: skip pos 2]  ; #{D800}
					57343 < w [pos: skip pos 2]  ; #{DFFF}
					56319 < w [break/return pos]  ; #{DBFF}
					none? lo: pick pos 3 [break/return pos]
					none? hi: pick pos 4 [break/return pos]
					56320 > w: hi * 256 + lo [break/return pos]  ; #{DC00}
					57343 >= w [pos: skip pos 4]  ; #{DFFF}
				]
				none
			] ; none = valid, break/return pos = invalid
		]
		32 [ ; UTF-32BE
			pos: data
			while [not tail? pos] [
				if any [
					4 > length? pos
					negative? c: to-integer pos
					1114111 < c  ; to-integer #{10FFFF}
				] [break/return pos]
			]
		]
		-32 [ ; UTF-32LE
			pos: data
			while [not tail? pos] [
				if any [
					4 > length? pos
					negative? c: also to-integer reverse/part pos 4 reverse/part pos 4
					1114111 < c  ; to-integer #{10FFFF}
				] [break/return pos]
			]
		]
	] [
		throw-error 'script 'invalid-arg num
	]
]
; Note: Native in R3, which doesn't support or screen the /utf option yet.
; See http://en.wikipedia.org/wiki/Unicode for charset/value explanations.

deline: funct [
	"Converts string terminators to standard format, e.g. CRLF to LF. (Modifies)"
	string [any-string!]
	/lines "Convert to block of lines (does not modify)"
] compose [
	linechar: (complement charset crlf)
	also case [
		not lines [ ; Convert the string in place
			; Change the newlines from any platform-specific line ending
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			parse/all/case either binary? string [as-string string] [string] [
				any [to cr a: cr opt lf b: (
					b: change/part a lf b
				) :b] to end
			]
			string
		] ; Otherwise generate a block of lines
		empty? string [copy []] ; Shortcut return for empty string
		'else [
			output: make block! divide length? string 50 ; Seems OK
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			if binary? string [string: as-string string]
			parse/all/case string [any [
				; Lines with stuff in them
				copy a some linechar (output: insert output a) [crlf | cr | lf | end]
				|
				; Empty lines
				[crlf | cr | lf] (output: insert output copy "")
			]]
			head output
		]
	] set [string output a b] none ; Cleanup after return
] ; /lines needs testing with final line ending once it works in R3.
; Note: Native in R3.

enline: funct [
	"Converts standard string terminators to current OS format, e.g. LF to CRLF. (Modifies)"
	series [any-string! block!]
	/with "Specifies alternate line termination."
	end-of-line [char! string!]
] compose [
	; The platform-specific line ending
	platform-line: (switch/default system/version/4 [
		3 15 [crlf] 2 [to-string either system/version/5 < 4 [cr] [lf]]
	] [to-string lf]) ; Don't know about Amiga (1) and Tao Elate (27)
	switch type?/word end-of-line [
		none! [end-of-line: platform-line]
		char! [end-of-line: to-string end-of-line]
	]
	either block? series [
		; First pass: Precalculate the length of the output string
		len: 0 len-eol: length? end-of-line
		foreach s series [len: len + len-eol + length? s]
		; Second pass: Build and fill the output string
		output: make string! len
		foreach s series [insert insert tail output s end-of-line]
		; Return and clean up
		also output set [series output] none
	] [
		unless end-of-line = "^/" [
			; Change the newlines to the platform-specific line ending
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			parse/all/case either binary? series [as-string series] [series] [
				any [to lf a: lf (a: change/part a end-of-line 1) :a] to end
			]
		]
		; Return and clean up
		also series set [series a] none
	]
] ; block! needs testing with final line ending once it works in R3.
; Note: Native in R3.
