REBOL [
	Title:  "REBOL Protocols: Utilities"
	Author: ["Carl Sassenrath" "Sterling Newton" "Jeff Kreis" "Holger Kruse"]
	Rights: "Copyright REBOL Technologies 1997-2002"
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

net-error: func [info] [make error! info] ; this func once had a THROW in it - which did not give correct error to SEND if no user/email field was set.

net-utils: make object! [

	confirm: func [
		"Confirm pairs of stimuli to their expected responses"
		port  [port!]  "Port to send/receive from"
		pairs [block! none!] "A block of pairs to send and expect"
		/multiline
		; needed by ESMTP for EHLO response
		/all "Return all the lines of multiline response"
		/local server-said an-error num server-said-lines
	][
		if found? pairs [
			an-error: [net-error reform ["Server error:" port/scheme server-said]]
			bind pairs 'port
			server-said-lines: make block! 3
			foreach [stimuli response] pairs [
				net-log [stimuli response]
				if stimuli <> 'none [insert port reform stimuli] ; sep w/space
				if response <> 'none [
					if none? server-said: pick port 1 [
						server-said: "connection failed"
						do an-error
					]
					append server-said-lines server-said
					if multiline [
						if ( pick server-said 4 ) = #"-" [
							num: copy/part server-said 3
							append num #" "
							until [
								net-log server-said
								server-said: pick port 1
								append server-said-lines server-said
								find/match server-said num
							]
						]
					]
					either block? response [
						if not foreach item response [
							if find/match/any server-said item [break/return true]
						] an-error
					][
						if not find/match/any server-said response an-error
					]
				]
				net-log server-said
			]
		]
		either all [server-said-lines] [server-said]
	]

	net-log: func [vals] [
		if net-watch [prin "Net-log: " probe reduce vals]
		vals
	]

	net-install: func [ 'proto-name proto-object proto-port-id /local blk x ][
		blk: [
			x: make system/standard/port [
				scheme: proto-name
				port-id: proto-port-id
				handler: proto-object
				passive: none ; used only in FTP 
				cache-size: 5 ;   right now (9-Aug-2000)
				proxy: make object! [
					host: port-id: user: pass: type: none bypass: none
				]
			]
		]
		change blk make set-word! proto-name
		system/schemes: make system/schemes blk
		if not system/options/quiet [print [proto-name "protocol loaded"] ]
	]

	URL-Parser: make object! [
		"First draft URL parser"

		scheme: user: pass: host: port-id: path: target: tag: p2: none
		vars: [user pass host port-id path target]

		;-- URL Character Sets:
		digit:       charset "0123456789"
		alpha-num:   charset [#"a" - #"z" #"A" - #"Z" #"0" - #"9"]
		scheme-char: insert copy alpha-num "+-."
		path-char:   insert copy alpha-num "=+-_.;:&$@%*',~?| []()^"!" ; !!! note: space allowed
		user-char:   insert copy alpha-num "=+-_.;&$%*,'#|"
		pass-char:   complement charset "^/ ^-@"
		;--- missing: encoded chars! !!!

		;-- URL Grammar:
		url-rules:   [scheme-part user-part host-part path-part file-part tag-part]
		scheme-part: [copy scheme some scheme-char #":" ["//" | none]]
		user-part:   [copy user uchars [#":" pass-part | none] #"@" | none (user: pass: none)]
		pass-part:   [copy pass to #"@" [skip copy p2 to "@" (append append pass "@" p2) | none]]
		host-part:   [copy host uchars [#":" copy port-id digits | none]]
		path-part:   [slash copy path path-node | none]
		path-node:   [pchars slash path-node | none]
		file-part:   [copy target pchars | none]
		tag-part:    [#"#" copy tag pchars | none]
		uchars:      [some user-char | none]
		pchars:      [some path-char | none]
		digits:      [1 5 digit]

		;-- Parse Function:
		parse-url: func [
			"Return url dataset or cause an error if not a valid URL"
			port [object! port!]
			url
			/set-scheme "Set port scheme"
			/no-error "Do not throw an error, return NONE instead."
		][  
			set vars none
			either parse/all url url-rules [
				if net-watch [print ["URL Parse:" reduce vars]]
				;-- Does not overwrite existing port values if new value not provided.
				if user [ port/user: user ]
				if pass [ port/pass: pass ]
				if host [ port/host: host ]
				if port-id [ port/port-id: to-integer port-id ]
				if path [ port/path: path ]
				if target [ port/target: target ]
				if all [set-scheme scheme] [port/scheme: to-word scheme]
				port
			][if not no-error [net-error reform ["URL error:" url]]]
		]

		set 'decode-url func [
			"Decode a URL into an object."
			url
			/local purl
		][
			purl: context [user: pass: host: port-id: path: target: none]
			parse-url/no-error purl url
		]
	]
	
	export: func [
		"Export an object to something that looks like a header"
		; Needs more work!
		object [object!] "Object to export"
		/local words values result word
	][
		;not truly correct, but should do for now
		words:  next first object    ; skip self
		values: next second object   ; skip self
		result: make string! (20 * length? words) ; a guess
		foreach word words [
			if found? first values [
				insert tail result reduce [word ": " first values newline]
			]
			values: next values
		]
		result
	]       

	 connect-proxy-socks4: func [port /sub-protocol subproto /local data sub-port check-return] [
		check-return: func [err] [
			if any [((length? data) < 2) ((second data) <> 90)] [
				system/words/close port
				net-error err
			]
		]
		if not sub-protocol [subproto: 'tcp]
		sub-port: system/words/open/lines compose [
			scheme: (to-lit-word subproto)
			host: port/proxy/host
			user: get in port/proxy 'user
			pass: get in port/proxy 'pass
			port-id: port/proxy/port-id
		]
		; 5.1(2).0.1(3) is SOCKS5, connect (bind) type connection, reserved, ipv4 address (hostname) coming

		either tuple? load port/host [
			hhh: load port/host
		] [
			hhh: read join dns:// port/host
		]
		write-io sub-port data: rejoin [
			#{0401}
			to-char port/port-id / 256
			to-char remainder port/port-id 256
			to-char first hhh
			to-char second hhh
			to-char third hhh
			to-char fourth hhh
			to-char 0
		] length? data
		read-io sub-port data: make binary! 8 8
		check-return reform ["Connection to" sub-port/host "failed"]
		sub-port
	]

	accept-proxy: func [dataport /local data] [
		check-return: func [err] [
			if any [((length? data) < 2) ((second data) <> 0)] [
				system/words/close dataport
				net-error err
			]
		]
		read-io dataport data: make binary! 10 10
		check-return reform ["Accepting connection from proxy failed"]
		data: at data 5
		dataport/host: copy/part data 4
		dataport/port-id: data/5 * 256 + data/6
	]

	connect-proxy: func [port mode /sub-protocol subproto /local data sub-port check-return host] [
		if not sub-protocol [subproto: 'tcp]
		if port/proxy/type = 'socks4 [return connect-proxy-socks4/sub-protocol port subproto]
		check-return: func [err] [
			if any [((length? data) < 2) ((second data) <> 0)] [
				system/words/close port
				net-error err
			]
		]
		sub-port: system/words/open/lines compose [
			scheme: (to-lit-word subproto)
			host: port/proxy/host
			user: get in port/proxy 'user
			pass: get in port/proxy 'pass
			port-id: port/proxy/port-id
		]
		; 5.2.0.2 is SOCKS V5, 2 choices, no auth and user/pass auth

		write-io sub-port #{05020002} 4
		read-io sub-port data: make binary! 2 2
		if any [((length? data) <> 2) ((second data) = 255)] [ ; rejected by the proxy server?
			system/words/close sub-port
			net-error ["Failed proxy handshake"]
		]
		if (second data) <> 0 [
			; sending: user/pass auth code, length of username, username, length of password, password
			write-io sub-port data: rejoin [
				(to-char 1)
				(to-char length? port/proxy/user)
				port/proxy/user
				(to-char length? port/proxy/pass)
				port/proxy/pass
			] length? data
			read-io sub-port data: make binary! 2 2
			check-return ["Failed proxy user authentication"]
		]
		; 5.1(2).0.1(3) is SOCKS5, connect (bind) type connection, reserved, ipv4 address (hostname) coming

		either (tuple? port/host) [
			host: join "" port/host
		] [
			host: port/host
		]
		write-io sub-port data: rejoin [
			#{05}
			(to-char either (mode == 'bind) [2] [1])
			(to-char 0)
			(to-char 3)
			(to-char length? host)
			host
			(to-char to-integer port/port-id / 256)
			(to-char to-integer remainder port/port-id 256)
		] length? data
		read-io sub-port data: make binary! 10 10
		check-return reform ["Connection to" sub-port/host "failed"]
		if (mode = 'bind) [
			data: at data 5
			sub-port/host: to-tuple copy/part data 4
			sub-port/port-id: data/5 * 256 + data/6
		]
		sub-port
	]
]

; RAMBO #3622
; Optional features (pad day and precise time in to-idate) are commented out.

to-itime: func [
	{Returns a standard internet time string (two digits for each segment)}
	time [time! number! block! none!]
	/local pad
][
	time: make time! time
	pad: func [n] [head insert/dup n: form n #"0" 2 - length? n]
	rejoin [
		pad time/hour ":" pad time/minute ":"
		pad round/down time/second
	]
]

to-idate: func [
	"Returns a standard Internet date string."
	date [date!]
	;/precise
	/local str ;pad
][
	;pad: func [n] [head insert/dup n: form n #"0" 2 - length? n]
	str: form date/zone
	remove find str ":"
	if (first str) <> #"-" [insert str #"+"]
	if (length? str) <= 4 [insert next str #"0"]
	reform [
		pick ["Mon," "Tue," "Wed," "Thu," "Fri," "Sat," "Sun,"] date/weekday
		;pad date/day
		date/day
		pick ["Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul" "Aug" "Sep" "Oct" "Nov" "Dec"] date/month
		date/year
		; fixed error when date/time was none
		to-itime any [date/time 0:00]
		;either precise [to-itime/precise date/time] [to-itime date/time]
		str
	]
]
