REBOL [
	Title:  "REBOL/View: Functions"
	Author: "Carl Sassenrath"
	Rights: "Copyright 2000 REBOL Technologies. All rights reserved."
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

viewed?: func [
	"Returns TRUE if face is displayed."
	face [object!]
][
	found? find system/view/screen-face/pane find-window face
]

insert-event-func: func [
	"Add a function to monitor global events. Return the func."
	funct [block! function!] "A function or a function body block"
][
	if block? :funct [funct: func [face event] funct]
	insert system/view/screen-face/feel/event-funcs :funct
	:funct
]

remove-event-func: func [
	"Remove an event function previously added."
	funct [function!]
][
	remove find system/view/screen-face/feel/event-funcs :funct
]

find-window: func [
	"Find a face's window face."  ; !!! Should window parent be none or screen-face?
	face [object!]
][
	while [face/parent-face][face: face/parent-face]
	face
]

in-window?: func [
	"Return true if a window contains a given face."
	window face
][
	if face = window [return true]
	if none? face/parent-face [return false]
	in-window? window face/parent-face
]

inform: func [
	"Display an exclusive focus panel for alerts, dialogs, and requestors."
	panel [object!]
	/offset where [pair!] "Offset of panel" 
	/title ttl [string!] "Dialog window title"
	/timeout time
][
	panel/text: copy any [ttl "Dialog"]
;   if none? panel/options [panel/options: copy [no-title]]
;   panel/edge: make face/edge [color: 200.200.200 size: 2x2 effect: 'bevel]
	panel/offset: either offset [where][system/view/screen-face/size - panel/size / 2]
	;if none? panel/feel [panel/feel: system/view/window-feel]
	panel/feel: system/view/window-feel
	show-popup panel
	; fix timeout: did not close the popup. now it calls hide-popup
	; if wait returned because of the timeout
	either time [if none? wait time [hide-popup/timeout]][do-events]
]

scroll-para: func [
	"Scroll a text face, given a scroller/slider face."
	tf "Text face to scroll"
	sf "Scroller face"
	/local tmp
][
	if none? tf/para [exit]
	tmp: tf/para/margin + tf/para/origin + size-text tf
	tmp: min 0x0 tf/size - tmp - either tf/edge [2 * tf/edge/size][0x0]
	either sf/size/x > sf/size/y [tf/para/scroll/x: sf/data * first tmp] [
		tf/para/scroll/y: sf/data * second tmp]
	show tf
]

dump-pane: dump-face: func [
	"Print face info for entire pane. (for debugging)"
	face [object!]
	/local depth style
][
	depth: ""  ;non-local data
	print [depth "Style:" all [in face 'style face/style] "Offset:" face/offset "Size:" face/size
		"Text:" if face/text [copy/part form face/text 20]]
	insert depth tab
	if object? face/pane [dump-face face/pane]
	if block? face/pane [foreach f face/pane [dump-face f]]
	remove depth
]

center-face: func [
	"Center a face relative to parent (screen for windows) or a sibling face."
	obj [object!]
	/with "Center relative to a sibling face." face [object!]
][
	if none? face [face: any [obj/parent-face system/view/screen-face]]
	obj/offset: max 0x0 face/size - obj/size / 2 + either with [face/offset] [0x0]
	obj
]

flag-face: func [
	"Sets a flag in a VID face."
	face [object!]
	'flag
][
	if none? face/flags [face/flags: copy [flags]]
	if not find face/flags 'flags [face/flags: copy face/flags insert face/flags 'flags]
	append face/flags flag
]

deflag-face: func [
	"Clears a flag in a VID face."
	face [object!]
	'flag
][
	if none? face/flags [exit]
	if not find face/flags 'flags [face/flags: copy face/flags insert face/flags 'flags]
	remove find face/flags flag
]

flag-face?: func [
	"Checks a flag in a VID face."
	face [object!] 'flag
] [all [face/flags find face/flags flag]]

clear-fields: func [
	"Clear all text fields faces of a layout."
	panel [object!]
][
	if not all [in panel 'type panel/type = 'face] [exit]
	unfocus
	foreach face panel/pane [
		if all [any-string? face/text flag-face? face tabbed] [
			clear face/text
			face/line-list: none
		]
	]
]

read-net: func [
	"Read a file from the net (web). Update progress bar. Allow abort."
	url [url!]
	/progress callback "Call func [total bytes] during transfer. Return true."
	/local port buffer data size

][
	vbug ['read-net url]
	if error? try [port: open/direct url][return none]
	size: to-integer any [port/locals/headers/content-length 8000]
	buffer: make binary! size
	set-modes port/sub-port [lines: false binary: true no-wait: true]
	until [
		if not data: wait [60 port/sub-port][data: true break]
		if data: copy port/sub-port [append buffer data]
		all [:callback size  not callback size length? buffer  data: true break]
		not data
	]
	close port
	if not data [buffer] ; return buffer or none if timeout
]

path-thru: func [
	"Return a path relative to the disk cache."
	url /local purl
][
	if file? url [return url]
	if not all [purl: decode-url url purl/host] [return none]
	rejoin [view-root/public slash purl/host slash any [purl/path ""] any [purl/target ""]]
]

exists-thru?: func [
	"Checks if a file is in the disk cache. Returns: none, false (out of date), or file"
	url [url! file!]
	/check "Checks version, checksum/secure, date, and/or size of cached file."
		info [block! tuple! binary! logic! date! none!]
	/local file data date
][
	if not file: path-thru url [return none]
	if not exists? file [return false]
	if not check [return file]
	if none? info [return file]
	if tuple? info [
		; if we cannot read it, then return NONE as if it does not exist.
		if not data: attempt [load/header script? file] [return none]
		if all [ ; returns true only if there is a header and its version >= requested version
			object? data: data/1 ; header
			tuple? data/version
			data/version >= info
		][return file] ; it is valid
		return false ; not valid
	]
	if binary? info [
		; if we cannot read it, then return NONE as if it does not exist.
		if not data: attempt [read/binary file] [return none]
		if info = checksum/secure data [return file] ; it is valid
		return false ; not valid
	]

	; If file exists, but info is bogus, return file (true). 
	if not block? info [info: reduce [info]]
	foreach item info [
		if date? item [
			date: get-modes file 'modification-date
			if (date + 0:0:3) < item [return false] ; Fudge is for windows FAT volumes
			;   (date - 0:0:3) > info/2  -- removed: do not overwrite local edits
		]
		if integer? item [
			if item <> size? file [return false]
		]
	]
	file
]

read-thru: func [
	"Read a net file from thru the disk cache. Returns binary, else none on error."
	url [url! file!]
	/progress callback "Call func [total bytes] during transfer.  Return true."
	/update "Force update from source site"
	/expand "Auto-decompress after transfer."
	/check "Update only if version, checksum/secure, or date/size do not match." info
	/to "Specify a file target, not cache." local-file [file! none!]
	/local file data purl loc-path
][
	vbug ['read-thru url info update]
	if none? file: path-thru url [return none]
	if local-file [file: local-file]
	if all [not update  exists-thru?/check file info] [
		if error? try [data: read/binary file] [return none]
		return data
	]
	if file? url [
		return attempt [read/binary url]
	]
	loc-path: first split-path file
	;print "reading............"
	if data: read-net/progress url :callback [
		if not exists? loc-path [
			if error? try [make-dir/deep loc-path] [return none]
		]
		if all [expand find/match data "rebpress"] [
			if error? try [data: decompress skip data 8][return none]
		]
		write/binary file data
		if all [check block? info info/2] [
			error? try [set-modes file [modification-date: info/2]]
		]
	]
	vbug ['read-thru-ok? found? data]
	data
]

load-thru: func [
	"Load a net file from the disk cache."
	url [url! file!]
	/update "Force update from source site"
	/binary "Force binary mode, otherwise expects REBOL data or code."
	/to "Specify local file target." local-file [file! none!] 
	/all "Return block with REBOL script including unevaluated header (safe mode)."
	/expand "Auto-decompress after transfer."
	/check "Update only if version, checksum/secure, or date/size do not match." info
	/local data
][
	if not data: either update [
		read-thru/update/to/expand url local-file
	][
		read-thru/to/expand/check url local-file info
	][
		return none
	]
	if binary [return load data]
	if error? try [
		data: either all [load/all to-string data][load to-string data] ;all is a refinement here
	][return none]
	data
]

do-thru: func [
	"Do a net file from the disk cache."
	url [url! file!]
	/args arg
	/update "Force update from source site"
	/check "Update only if version, checksum/secure, or date/size do not match." info
	/boot "Run as boot script (sets system/options/path)"
	/local data
][
	if not data: either update [
		read-thru/update/expand url
	][
		read-thru/check/expand url info
	][
		alert reform ["Cannot run" url]
		return none
	]
	url: path-thru url
	if boot [system/options/path: first split-path url]
	do/args url arg
]

launch-thru: func [
	"Launch a net file from the disk cache."
	url [url! file!]
	/update "Force update from source site"
	/check "Update only if version, checksum/secure, or date/size do not match." info
	/local data
][
	if not data: either update [
		read-thru/update/expand url
	][
		read-thru/check/expand url info
	][
		alert reform ["Cannot launch" url]
		return none
	]
	launch/secure-cmd path-thru url
]

set 'load-image func [
	"Load an image through an in-memory image cache."
	image-file [file! url!] "Local file or remote URL"
	/update "Force update from source site"
	/clear "Purge the entire cache"
	/local image image-cache
][
	image-cache: []  ; static block
	if clear [system/words/clear image-cache recycle]
	if any [update  not image: select image-cache image-file] [
		if all [update image] [remove/part find image-cache image-file 2]
		repend image-cache [
			image-file
			image: either file? image-file [load image-file][
				either update [load-thru/binary/update image-file][load-thru/binary image-file]
			]
		]
	]
	image
]

request-download: func [
	"Request a file download from the net. Show progress. Return none on error."
	url [url!]
	/to "Specify local file target." local-file [file! none!] 
	/local prog lo stop data stat event-port event
][
	view/new center-face lo: layout [
		backeffect [gradient 1x1 water gray]
		space 10x8
		vh2 300 gold "Downloading File:"
		vtext bold center 300 to-string url
		prog: progress 300
		across
		btn 90 "Cancel" [stop: true]
		stat: text 160x24 middle
	]
	stop: false
	data: read-thru/to/progress/update url local-file func [total bytes][
		prog/data: bytes / (max 1 total)
		stat/text: reform [bytes "bytes"]
		show [prog stat]
		not stop
	]
	unview/only lo
	if not stop [data]
]

vbug: func [msg [block! any-string!] /local dbg] [
	if not dbg: user-prefs/debug [exit]
	msg: head insert remold msg "view "
	either file? dbg [write/append dbg msg][print msg]
]
